package edu.fiu.jtlex.algorithms;

import org.jacop.core.IntVar;
import org.jacop.core.Store;
import org.jacop.search.*;

import edu.fiu.jtlex.timeml.graph.ITimeMLGraphTCSP;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

/**
 * Solver of a TCSP that is created from a graph.
 *
 * @since jTLEX 1.0
 */
public class GraphTCSPSolver implements IGraphTCSPSolver {

    /**
     * Creates a new instance
     */
    public GraphTCSPSolver() {}

    /**
     * Solves the given problem using minimum selection strategy.
     *
     * @param problem The problem to solve
     * @return The solution to the problem
     * @throws NullPointerException If the problem is null
     * @since jTLEX 1.0
     */
    @Override
    public List<IntVar> solve(ITimeMLGraphTCSP problem) {
        Objects.requireNonNull(problem);
        return solve(problem, new IndomainMin<>());
    }

    /**
     * Solves a problem using random selection strategy.
     *
     * @param problem The problem to solve
     * @return The solution to the problem
     * @throws NullPointerException If the problem is null
     * @since jTLEX 1.0
     */
    @Override
    public List<IntVar> solveRandom(ITimeMLGraphTCSP problem) {
        Objects.requireNonNull(problem);
        return solve(problem, new IndomainRandom<>());
    }

    /**
     * Solves the given TCSP using the given selection strategy in a 10
     * seconds time window.
     *
     * @param problem The problem to solve
     * @param strategy The selection strategy to use
     * @return The solution to the problem
     * @since jTLEX 1.0
     */
    private List<IntVar> solve(ITimeMLGraphTCSP problem, Indomain<IntVar> strategy) {

        Search<IntVar> search = new DepthFirstSearch<>();
        search.setTimeOut(10);
        search.setPrintInfo(false);

        Store store = problem.getStore();
        IntVar[] timePoints = problem.getTimepoints();

        SelectChoicePoint<IntVar> select = new InputOrderSelect<>(
                store, timePoints, strategy
        );

        boolean solved = search.labeling(store, select);
        if (!solved) return null;

        List<IntVar> solution = new ArrayList<>();
        Collections.addAll(solution, timePoints);
        solution.sort(this::compareIntVar);
        return solution;
    }

    /**
     * Comparator for IntVar objects based on their assigned integer value.
     *
     * @param TP1 An IntVar object.
     * @param TP2 An IntVar object.
     * @return If TP1 is after TP2 then +1, if TP1 is before TP2 then -1, 0
     * otherwise.
     * @throws NullPointerException If either time point is null
     * @since jTLEX 1.0
     */
    private int compareIntVar(IntVar TP1, IntVar TP2) {
        Objects.requireNonNull(TP1);
        Objects.requireNonNull(TP2);
        return Integer.compare(TP1.value(), TP2.value());
    }

    /**
     * Converts the given solution into its string representation
     *
     * @param solution The solution to convert
     * @return A string representation of the solution
     * @throws NullPointerException If the solution is null
     * @since jTLEX 1.0
     */
    @Override
    public List<String> convertSolutionToString(List<IntVar> solution) {
        Objects.requireNonNull(solution);
        return solution.stream().map(IntVar::toString).toList();
    }
}
