package edu.fiu.jtlex.algorithms;

import java.util.List;
import java.util.Set;

import edu.fiu.jtlex.timeml.graph.ITimeMLGraph;
import edu.fiu.jtlex.timeml.timeline.ITimeMLTimeline;

/**
* Represents the main processing area for jTLEX.
* Creates different parsers/detectors for a graph to utilize.
*
* @since jTLEX 1.0
*/
public interface ITLEX {
	/**
	 * Returns the set of partitions within a graph.
	 * 
	 * @since jTLEX 1.0
	 * @return the a List of IGraphs containing the subgraphs.
	 */
    List<ITimeMLGraph> getPartitions();

	/**
	 * Returns the set of consistent subordinate subgraphs.
	 * 
	 * @since jTLEX 1.0
	 * @return the set of consistent subordinate subgraphs.
	 */
    List<ITimeMLGraph> getConsistentPartitions();

	/**
	 * Returns the set of inconsistent subordinate subgraphs.
	 * 
	 * @since jTLEX 1.0
	 * @return the set of inconsistent subordinate subgraphs.
	 */
    List<ITimeMLGraph> getInconsistentPartitions();
    
	/**
	 * Gets the set of inconsistent subgraphs
	 * 
	 * @return The set of inconsistent partitions
	 * @since jTLEX 1.0
	 */
	Set<ITimeMLGraph> getInconsistentSubGraphs();

	/**
	 * Return a list of solutions for the graph.
	 * 
	 * @since jTLEX 1.0
	 * @return the list of list of string representing solutions.
	 */
    List<List<String>> getSolutions();

	/**
	 * Returns the timeline of the graph.
	 * 
	 * @since jTLEX 1.0
	 * @return an ITimeline object containing the timeline of the graph.
	 **/
    ITimeMLTimeline getTimeline();
    
	/**
	 * Returns the length of the timeline.
	 * 
	 * @since jTLEX 1.0
	 * @return the length of the main timeline, -1 if graph is inconsistent.
	 */
    int getTimelineLength();

	/**
	 * Returns a list of Indeterminant Time Pairs.
	 * 
	 * @since jTLEX 1.0
	 * @return a list of strings containing the Indeterminant Time Pairs.
	 */
    List<String> getIndeterminantTimePairs();

	/**
	 * Returns a percent value representing the indeterminacy score.
	 * 
	 * @since jTLEX 1.0
	 * @return a double that is the percent value representing the indeterminacy score.
	 */
    double getIndeterminacyScore();

	/**
	 * Checks if the graph is consistent.
	 * 
	 * @since jTLEX 1.0
	 * @return true if the graph is consistent, or else false.
	 */
    boolean isConsistent();

	/**
	 * Produces a JSON representation of the object.
	 *
	 * @return A JSON string representation of the object.
	 * @since jTLEX 1.0
	 */
	String toJson();
}
