package edu.fiu.jtlex.timeml.graph;

import java.util.List;
import java.util.Objects;
import java.util.Set;

import edu.fiu.jtlex.timeml.data.ITimeMLLink;

/**
 * Default implementation of IPartitionResult.
 *
 * @since jTLEX 1.0
 */
public class PartitionResult implements IPartitionResult {

    private final List<ITimeMLGraph> _partitions;
    private final Set<ITimeMLLink> _partitionLinks;

    /**
     * Creates a new partition result.
     *
     * @param partitions List of the partitions.
     * @param partitionLinks Set of the single S-links.
     * @param timeMLGraph the main/parent timeMLGraph
     * @throws NullPointerException If any of the arguments is null.
     * @since jTLEX 1.0
     */
    public PartitionResult(List<ITimeMLGraph> partitions, Set<ITimeMLLink> partitionLinks, ITimeMLGraph timeMLGraph) {
        Objects.requireNonNull(partitions);
        Objects.requireNonNull(partitionLinks);

        if (!partitions.isEmpty()) {
            sortAndClassify(partitions, timeMLGraph);
        }

        _partitions = partitions;
        _partitionLinks = partitionLinks;
    }

    /**
     * Sorts the partitions by decreasing order in node counts.
     * That is, the largest partition first.
     * Then classifies them.
     * The largest partition is the main subgraph, the rest are subordination graphs.
     *
     * @param partitions List of partitions to sort and classify
     * @since jTLEX 1.0
     */
    private void sortAndClassify(List<ITimeMLGraph> partitions, ITimeMLGraph timeMLGraph) {
        partitions.sort((g1, g2) -> Integer.compare(g2.getNodes().size(), g1.getNodes().size()));

        for (int i = 0; i < partitions.size(); i++) {
        	
        	ITimeMLGraph subGraph = timeMLGraph.subGraph(partitions.get(i).getNodes(), partitions.get(i).getLinks(), partitions.get(i).getEvents());
            partitions.set(i, subGraph);
        }
    }

    /**
     * Gets the list of partitions.
     *
     * @return The list of partitions.
     * @since jTLEX 1.0
     */
    @Override
    public List<ITimeMLGraph> getPartitions() {
        return _partitions;
    }

    /**
     * Gets the set of partition links.
     *
     * @return The set of partition links.
     * @since jTLEX 1.0
     */
    @Override
    public Set<ITimeMLLink> getPartitionLinks() {
        return _partitionLinks;
    }

}
