package edu.fiu.jtlex.timeml.graph;

import org.jacop.constraints.PrimitiveConstraint;
import org.jacop.constraints.XeqY;
import org.jacop.constraints.XgtY;
import org.jacop.constraints.XltY;
import org.jacop.core.IntVar;
import org.jacop.core.Store;

import edu.fiu.jtlex.timeml.data.ITimeMLLink;

import edu.fiu.jtlex.timeml.data.ITimeMLLink.ITimeMLLinkType;

import java.util.*;

/**
 * Implements a TimeML Graph as a Temporal Constraint Satisfaction Problem.
 * Also acts as an abstraction for creating Stores.
 *
 * @see ITimeMLGraphTCSP
 * @since jTLEX 1.0
 */
public class TimeMLGraphTCSP implements ITimeMLGraphTCSP {

    private final IntVar[] _timePoints;
    private final Store _store;

    /**
     * Transforms the TimeML Graph into a Temporal Constraint Satisfaction Problem.
     * This method created two JaCoP IntVars per node, one for the start point, the other
     * for the end point, and asserts constraints depending on the types of links in the graph.
     *
     * @param timeMlGraph A partitioned timeML graph to convert into a TCSP
     * @throws NullPointerException If the graph is null
     * @throws IllegalStateException If the graph contains any S-Link
     * @since jTLEX 1.0
     */
    public TimeMLGraphTCSP(ITimeMLGraph timeMlGraph) {

        Objects.requireNonNull(timeMlGraph);
        
        if (containsSLink(timeMlGraph.getLinks())) {  	
            throw new IllegalStateException("The graph cannot contain any S-Link. Try partitioning before creating a TCSP");
    }
        // Initialize set of time points variables.
        int timePointsCount = 2 * timeMlGraph.getNodes().size();
        _timePoints = new IntVar[timePointsCount];

        List<ITimeMLNode> nodes = timeMlGraph.getNodes().stream().toList();
        List<ITimeMLLink> links = timeMlGraph.getLinks().stream().toList();

        // Map a node ID string to the node's timepoint index.
        Map<String, Integer> nodesIndexes = new HashMap<>();

        // Represents the TCSP including the variables and the constrains.
        _store = new Store();

        /*
         * From each node in the nodes set we get two discrete Integer variables
         * in the TCSP. Each variable can take a value from 1 to timePointsCount.
         */
        for (int i = 0; i < nodes.size(); i++) {
            // Get the String ID for this node. Ex: e1.
            String nodeID = nodes.get(i).getIdStr();

            // This node's time points are in position i and i+1 in the _timePoints array.
            nodesIndexes.put(nodeID, 2 * i);

            _timePoints[2 * i] = new IntVar(_store, nodeID + "-", 1, timePointsCount);
            _timePoints[(2 * i) + 1] = new IntVar(_store, nodeID + "+", 1, timePointsCount);

            // For each node n, its time points must satisfy that n- < n+.
            _store.impose(new XltY(_timePoints[2 * i], _timePoints[(2 * i) + 1]));
        }

        /*
         * For each link, generate a set of constrains (maximum of 2) for the TCSP.
         */
        for (ITimeMLLink link : links) {
            // Get the nodes involved in this link.
            ITimeMLNode nodeA = timeMlGraph.getNodeById(link.getEventInstance());
            ITimeMLNode nodeB = timeMlGraph.getNodeById(link.getRelatedToNode());

            // Get the link type.
            ITimeMLLinkType linkType = link.getRelType();

            // Using the IDs of these nodes get where they are in the _timePoints array.
            int nodeAIntVarIndex = nodesIndexes.get(nodeA.getIdStr());
            int nodeBIntVarIndex = nodesIndexes.get(nodeB.getIdStr());

            // Timepoints for node A.
            IntVar nodeAMinus = _timePoints[nodeAIntVarIndex];
            IntVar nodeAPlus = _timePoints[nodeAIntVarIndex + 1];

            // Time points for node B.
            IntVar nodeBMinus = _timePoints[nodeBIntVarIndex];
            IntVar nodeBPlus = _timePoints[nodeBIntVarIndex + 1];

            List<PrimitiveConstraint> constraints = generateConstraints(
                    linkType, nodeAMinus, nodeAPlus, nodeBMinus, nodeBPlus
            );

            for (PrimitiveConstraint constraint : constraints) {
                // Impose the constrain into the TCSP.
                _store.impose(constraint);
            }
        }
    }

    /**
     * Given a link type (ALink or TLink) this method generates a list constraints
     * to be enforced for the given link.
     *
     * @param linkType An ALinkType or TLinkType enum value.
     * @param nodeAMinus Start of first time point
     * @param nodeAPlus End of first time point
     * @param nodeBMinus Start of second time point
     * @param nodeBPlus End of second time point
     * @return A List with 1 or 2 PrimitiveConstraint objects.
     * @throws IllegalArgumentException If the link type is not A-Link nor T-Link.
     * @since jTLEX 1.0
     */
    private List<PrimitiveConstraint> generateConstraints(
            ITimeMLLinkType linkType, IntVar nodeAMinus, IntVar nodeAPlus, IntVar nodeBMinus, IntVar nodeBPlus) {

        if (linkType instanceof ITimeMLLink.TLink TLink) {
            return getTLinkConstraints(TLink, nodeAMinus, nodeAPlus, nodeBMinus, nodeBPlus);
        } else if (linkType instanceof ITimeMLLink.ALink ALink) {
            return getALinkConstraints(ALink, nodeAMinus, nodeAPlus, nodeBMinus, nodeBPlus);
        } else {
            throw new IllegalArgumentException();
        }
    }

    /**
     * Generates the constraints to be enforced to the given T-Link.
     *
     * @param TLink The T-Link to constrain
     * @param nodeAMinus Start of first time point
     * @param nodeAPlus End of first time point
     * @param nodeBMinus Start of second time point
     * @param nodeBPlus End of second time point
     * @return The List of constraints
     * @throws IllegalArgumentException If the T-Link is not specified in the enum
     * @since jTLEX 1.0
     */
    private List<PrimitiveConstraint> getTLinkConstraints(
            ITimeMLLink.TLink TLink, IntVar nodeAMinus, IntVar nodeAPlus, IntVar nodeBMinus, IntVar nodeBPlus) {

        List<PrimitiveConstraint> constraints = new ArrayList<>();

        constraints.add(new XltY(nodeAMinus, nodeAPlus));
        constraints.add(new XltY(nodeBMinus, nodeBPlus));

        switch (TLink) {
            case BEFORE:
                // Before TLink requires one constrains: A+ < B-.
                constraints.add(new XltY(nodeAPlus, nodeBMinus));
                break;
            case AFTER:
                // After TLink enforces A- > B+.
                constraints.add(new XgtY(nodeAMinus, nodeBPlus));
                break;
            case INCLUDES:
                // INCLUDES TLink requires two constrains: A- < B- AND B+ < A+.
                constraints.add(new XltY(nodeAMinus, nodeBMinus));
                constraints.add(new XltY(nodeBPlus, nodeAPlus));
                break;
            case IS_INCLUDED:
                // IS_INCLUDED TLink requires two constrains: B- < A- AND A+ < B+.
                constraints.add(new XltY(nodeBMinus, nodeAMinus));
                constraints.add(new XltY(nodeAPlus, nodeBPlus));
                break;
            case DURING:
                // DURING TLink requires two constrains: B- = A- AND A+ = B+.
                constraints.add(new XeqY(nodeBMinus, nodeAMinus));
                constraints.add(new XeqY(nodeAPlus, nodeBPlus));
                break;
            case DURING_INV:
                // DURING_INV TLink requires two constrains: A- = B- AND B+ = A+.
                constraints.add(new XeqY(nodeAMinus, nodeBMinus));
                constraints.add(new XeqY(nodeBPlus, nodeAPlus));
                break;
            case SIMULTANEOUS:
            case IDENTITY:
                // IDENTITY TLink requires two constraints: A- = B- AND A+ = B+.
                // SIMULTANEOUS TLink requires two constraints: A- = B- AND A+ = B+.
                constraints.add(new XeqY(nodeAMinus, nodeBMinus));
                constraints.add(new XeqY(nodeAPlus, nodeBPlus));
                break;
            case IAFTER:
                // IAFTER TLink requires one constraint: B+ = A-.
                constraints.add(new XeqY(nodeBPlus, nodeAMinus));
                break;
            case IBEFORE:
                // IBEFORE TLink requires one constraint: A+ = B-.
                constraints.add(new XeqY(nodeAPlus, nodeBMinus));
                break;
            case BEGINS:
                // BEGINS TLink requires two constraints: A- = B- AND A+ < B+.
                constraints.add(new XeqY(nodeAMinus, nodeBMinus));
                constraints.add(new XltY(nodeAPlus, nodeBPlus));
                break;
            case ENDS:
                // ENDS TLink requires two constraints: B- < A- AND A+ = B+.
                constraints.add(new XltY(nodeBMinus, nodeAMinus));
                constraints.add(new XeqY(nodeAPlus, nodeBPlus));
                break;
            case BEGUN_BY:
                // BEGUN_BY TLink requires two constraints: A- = B- AND B+ < A+.
                constraints.add(new XeqY(nodeAMinus, nodeBMinus));
                constraints.add(new XltY(nodeBPlus, nodeAPlus));
                break;
            case ENDED_BY:
                // ENDED_BY TLink requires two constraints: A- < B- AND A+ = B+.
                constraints.add(new XltY(nodeAMinus, nodeBMinus));
                constraints.add(new XeqY(nodeBPlus, nodeAPlus));
                break;
            default:
                throw new IllegalArgumentException();
        }

        return constraints;
    }

    /**
     * Generates the constraints to be enforced to the given A-Link.
     *
     * @param ALink The A-Link to constrain
     * @param nodeAMinus Start of first time point
     * @param nodeAPlus End of first time point
     * @param nodeBMinus Start of second time point
     * @param nodeBPlus End of second time point
     * @return The List of constraints
     * @throws IllegalArgumentException If the A-Link is not specified in the enum
     * @since jTLEX 1.0
     */
    private List<PrimitiveConstraint> getALinkConstraints(
            ITimeMLLink.ALink ALink, IntVar nodeAMinus, IntVar nodeAPlus, IntVar nodeBMinus, IntVar nodeBPlus) {

        List<PrimitiveConstraint> constraints = new ArrayList<>();

        constraints.add(new XltY(nodeAMinus, nodeAPlus));
        constraints.add(new XltY(nodeBMinus, nodeBPlus));

        switch (ALink) {
            case INITIATES:
                // INITIATES = BEGINS -> TLink requires two constraints: A- = B- AND A+ < B+.
                constraints.add(new XeqY(nodeAMinus, nodeBMinus));
                constraints.add(new XltY(nodeAPlus, nodeBPlus));
                break;
            case CULMINATES:
            case TERMINATES:
                // TERMINATES = ENDS -> TLink requires two constraints: B- < A- AND A+ = B+.
                // CULMINATES = ENDS -> TLink requires two constraints: B- < A- AND A+ = B+.
                constraints.add(new XltY(nodeBMinus, nodeAMinus));
                constraints.add(new XeqY(nodeAPlus, nodeBPlus));
                break;
            case CONTINUES:
            case REINITIATES:
                // REINITIATES = IS_INCLUDED -> TLink requires two constrains: B- < A- AND A+ < B+.
                // CONTINUES = IS_INCLUDED -> TLink requires two constrains: B- < A- AND A+ < B+.
                constraints.add(new XltY(nodeBMinus, nodeAMinus));
                constraints.add(new XltY(nodeAPlus, nodeBPlus));
                break;
            default:
                throw new IllegalArgumentException();
        }

        return constraints;
    }

    /**
     * Checks whether the given set of Links contains any S-Link.
     *
     * @param links The set to be scanned
     * @return True if the set contains any S-Link, else false.
     * @since jTLEX 1.0
     */
    private boolean containsSLink(Set<ITimeMLLink> links) {
        return links.stream().anyMatch(l -> l.getRelType() instanceof ITimeMLLink.SLink);
    }

    @Override
    public Store getStore() {
        return _store;
    }

    @Override
    public IntVar[] getTimepoints() {
        return _timePoints;
    }
}
