package edu.fiu.jtlex.timeml.timeline;

import java.util.ArrayList;

import java.util.Iterator;
import java.util.List;
import java.util.Objects;
import java.util.Set;

import edu.fiu.jtlex.timeml.data.ITimeMLLink;
/**
 *
 * Default implementation of {@code ITimeline}.
 * 
 * @author Emmanuel Garcia
 * @since jTLEX 1.0
 */
public class TimeMLTimeline implements ITimeMLTimeline, Iterator<String>, Iterable<String>{
	
	private List<String> MainTimeline;
	private List<String> attachmentPoints;
	private List<List<String>> Timelines;
	private List<List<String>> SubordinationTimelines;
	private Set<ITimeMLLink> singleSlinks;
	private int currentPosition;
	private int totalTimepoints;

	/**
	 * Represents a Timeline of a TimeML graph, which is an ordered list of the events
	 * extracted from a TimeML Graph. TimeML annotations are converted into a collection of
	 * main and subordinated timelines arranged into a trunk-and-branch style structure.
	 * This class takes an ordered timeline, a set of attachment links, arranging them into a Timeline.
	 * @param Timeline produced by the Solve() method in the GraphTCSP.java class.
	 * @param singleSlinks {@code Set<ILink>} produced by the partitionLinks() method in the Partitioner.java class.
	 * @throws NullPointerException if Timeline argument is <code>null</code>
	 */
	public TimeMLTimeline(List<List<String>> Timeline, Set<ITimeMLLink> singleSlinks){
		this.Timelines = Timeline;
		this.singleSlinks = singleSlinks;
		totalTimepoints = 0;
		SubordinationTimelines = new ArrayList<>();
		countTimepoints();
		initializeMainTimeline();
		initializeSubordinationTimelines();
		if(singleSlinks != null)
			initializeAttachmentPoints();
		currentPosition = 0;
	}
	
	/**
	 * Returns an iterator object that is used in for-each loops.
	 * @return an iterator object that is used in for-each loops.
	 * @since jTLEX 1.0
	 */
	public Iterator<String> iterator(){
		// Must return to head after each loop
		this.currentPosition = 0;
		return this;
	}
	
	
	/**
	 * Checks if there are Timeline elements remaining in the list.
	 * @return true if there are Timeline elements remaining in the list.
	 * @since jTLEX 1.0
	 */
	public boolean hasNext() {
		if(currentPosition < MainTimeline.size()) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Returns next String representation of an event on the timeline.
	 * @return next String representation of an event on the timeline.
	 * @since jTLEX 1.0
	 */
	public String next() throws IndexOutOfBoundsException {
		return MainTimeline.get(currentPosition++);
	}
	
	/**
	 * Returns the {@code List<String>} representation of the main Timeline in the graph.
	 * @return the {@code List<String>} representation of the main Timeline in the graph.
	 * @since jTLEX 1.0
	 */
	public List<String> getMainTimeline() {
		return MainTimeline;
	}
	
	/**
	 * Returns an ArrayList containing the Timelines belonging to the subordinating subgraphs.
	 * @return an ArrayList containing the Timelines belonging to the subordinating subgraphs.
	 * @since jTLEX 1.0
	 */
	public List<List<String>> getSubordinationTimelines(){
		return SubordinationTimelines;
	}
	
	/**
	 * Returns number of timepoints.
	 * @return the number of timepoints.
	 * @since jTLEX 1.0
	 */
	public int getTotalTimePoints() {

		return totalTimepoints;
	}
	
	/**
	 * Returns a List containing the breaking points.
	 * @return a List containing the breaking points.
	 * @since jTLEX 1.0
	 */
	public List<String> getAttachmentPoints() {
		return attachmentPoints;
	}
	
	/**
	 * Returns the first timepoint in the Timeline.
	 * @return the first timepoint in the Timeline.
	 * @since jTLEX 1.0
	 */
	public String getFirstPoint() {
		return MainTimeline.get(0);
	}
	
	/**
	 * Returns the last timepoint in the Timeline.
	 * @return the last timepoint in the Timeline.
	 * @since jTLEX 1.0
	 */
	public String getLastPoint() {
		return MainTimeline.get(MainTimeline.size() - 1);
	}
	
	/**
	 * Returns the number of events in the timeline.
	 * @return the number of events in the timeline.
	 * @since jTLEX 1.0
	 */
	public int getTimelineLength() {
		//split the last time point in edge/number
		String[] time_point = MainTimeline.get(MainTimeline.size() - 1).split(" = ");
		return Integer.parseInt(time_point[1]);
	}
	
	/**
	 * Returns the total number of subordination branches in the graph.
	 * @return the total number of subordination branches in the graph.
	 * @since jTLEX 1.0
	 */
	public int getTotalSubordinationBranches() {
		return SubordinationTimelines.size();
	}
	
	/**
	 * Helper method to initialize the main timeline variable.
	 * @since jTLEX 1.0
	 */
	private void initializeMainTimeline() {
		MainTimeline = Timelines.get(0);
		for (List<String> list : Timelines) {
			int listLength = Integer.parseInt(list.get(list.size() - 1).split(" = ")[1]);
			int mainTimelineLegth = Integer.parseInt(MainTimeline.get(MainTimeline.size() - 1).split(" = ")[1]);
			
			if (listLength > mainTimelineLegth) {
				MainTimeline = list;
			}
		}
		Timelines.remove(MainTimeline);
	}
	
	/**
	 *  Helper method to initialize the subordination timelines.
	 * @since jTLEX 1.0
	 */
	private void initializeSubordinationTimelines() {
		if(!Timelines.isEmpty()) {
			for (List<String> list : Timelines) {
				SubordinationTimelines.add(list);
			}
		}
	}
	
	/**
	 * Helper method to initialize the attachment points
	 * @since jTLEX 1.0
	 */
	private void initializeAttachmentPoints() {
		attachmentPoints = new ArrayList<String>();
		if(!singleSlinks.isEmpty()) {
			for(ITimeMLLink link : singleSlinks) {
				attachmentPoints.add(link.getEventInstance()+"->"+link.getRelatedToNode());
			}
		}
	}
	
	/**
	 * Helper method to count the number of timepoints.
	 * @since jTLEX 1.0
	 */
	private void countTimepoints() {
		for (List<String> list : Timelines) {
			this.totalTimepoints += list.size();
		}
	}
	
	@Override
	public String toString() {
		String ret = "{";
		
		for (String item : this.MainTimeline) {
			ret += "\n\t" + item;
		}
		
		ret += "\n}";
		
		return ret;
	}

	
	/**
	 * Returns the JSON (RFC 8259) representation of the Timeline.
	 * @return the JSON (RFC 8259) representation of the Timeline.
	 * @since jTLEX 1.0
	 */
	public String toJson() {
        String out = "[";
        String first, id, eventBoundary, position;
        
        for (String timeline : this.MainTimeline) {
            out += "{";
            
            first = timeline.substring(0, timeline.indexOf("="));
            id = first.substring(0, first.length()-2);
            eventBoundary = first.substring(first.length()-2, first.length()-1);
            position = timeline.substring(timeline.indexOf("=") + 2, timeline.length());
            
            out += "\"id\":\"" + id + "\",";
            out += "\"eventBoundary\":\"" + eventBoundary + "\",";
            out += "\"position\":\"" + position + "\"";
            
            out += "},";
        }
        
        out = out.substring(0, out.length()-1) + "]";
        
        
        return out;
    }
	
	/**
	 * Generates a hash code for the timeline.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
	 */
    @Override
    public int hashCode() {
        int hashCode = 1;
        hashCode = 31*hashCode + (this.MainTimeline==null ? 0 : this.MainTimeline.hashCode());
        hashCode = 31*hashCode + (this.attachmentPoints==null ? 0 : this.attachmentPoints.hashCode());
        hashCode = 31*hashCode + (this.Timelines==null ? 0 : this.Timelines.hashCode());
        hashCode = 31*hashCode + (this.SubordinationTimelines==null ? 0 : this.SubordinationTimelines.hashCode());
        hashCode = 31*hashCode + (this.singleSlinks==null ? 0 : this.singleSlinks.hashCode());
        hashCode = 31*hashCode + currentPosition;
        hashCode = 31*hashCode + totalTimepoints;
        
        return hashCode;
    }
	
	/**
	 * Checks if this Timeline equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if timelines are equal, i.e. same mainTimeline and timelines, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o) {
        Objects.requireNonNull(o);

        if (!(o instanceof ITimeMLTimeline))
            return false;

        ITimeMLTimeline obj = (TimeMLTimeline) o;

        if (this.MainTimeline != obj.getMainTimeline())
            return false;
        
        if (this.attachmentPoints != obj.getAttachmentPoints())
        	return false;

        if (this.SubordinationTimelines != obj.getSubordinationTimelines())
        	return false;
        
        if (this.totalTimepoints != obj.getTotalTimePoints())
        	return false;
  

        return true;
    }
}
